/*
Copyright (c) 2021 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#undef _WIN32_WINNT
#define _WIN32_WINNT 0x0501
#undef UNICODE
#define UNICODE
#define OEMRESOURCE
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shellapi.h> // Linker: shell32
#ifdef _MSC_VER
# pragma warning(push)
# pragma warning(disable : 4706 4996 28159)
#endif

HCURSOR glob_defaultCur; // handle of the default arrow mouse pointer

// Playing with the appearance of the system cursor is scary. We have to ensure that we restore it, whatever happens!
static BOOL WINAPI RestoreDefaultCursor(DWORD unused)
{
  (void)unused;
  SetSystemCursor(glob_defaultCur, OCR_NORMAL);
  return FALSE;
}

// check whether the mouse is out of the work area rectangle of the display monitor, or it is on an auto-hide bar (like the task bar or a tool bar which behaves like a task bar)
static BOOL IsCursorOutOfWorkArea(const DWORD *const pAppBarMessage, const HWND hConWnd, const POINT *const pMousePos)
{
  const HMONITOR hMon = MonitorFromWindow(hConWnd, MONITOR_DEFAULTTONEAREST);
  MONITORINFO monInfo = { .cbSize = sizeof(MONITORINFO) };
  if (!hMon || !GetMonitorInfoW(hMon, &monInfo) || monInfo.rcWork.left >= pMousePos->x || monInfo.rcWork.top >= pMousePos->y || monInfo.rcWork.right <= pMousePos->x || monInfo.rcWork.bottom <= pMousePos->y)
    return TRUE; // mouse is out of the work area rectangle of the display monitor

  APPBARDATA barData = { .cbSize = sizeof(APPBARDATA), .rc = monInfo.rcMonitor };
  for (barData.uEdge = ABE_LEFT; barData.uEdge <= ABE_BOTTOM; ++barData.uEdge) // search auto-hide bars clockwise from the left edge to the bottom edge of the monitor
  {
    const HWND hBarWnd = (HWND)SHAppBarMessage(*pAppBarMessage, &barData);
    RECT rcBar;
    if (!hBarWnd || !GetWindowRect(hBarWnd, &rcBar))
      continue; // no auto-hide bar found on the current edge, continue searching on the next edge of the monitor

    switch ((int)barData.uEdge)
    {
    case ABE_LEFT:
      if (rcBar.right > monInfo.rcWork.left + (rcBar.right - rcBar.left) / 2)
        return TRUE; // auto-hide bar is visible which means the mouse hovers over it (same for the below case statements)
      break;
    case ABE_TOP:
      if (rcBar.bottom > monInfo.rcWork.top + (rcBar.bottom - rcBar.top) / 2)
        return TRUE;
      break;
    case ABE_RIGHT:
      if (rcBar.left < monInfo.rcWork.right - (rcBar.right - rcBar.left) / 2)
        return TRUE;
      break;
    default: // ABE_BOTTOM
      if (rcBar.top < monInfo.rcWork.bottom - (rcBar.bottom - rcBar.top) / 2)
        return TRUE;
    }
  }

  return FALSE;
}

// convert the mouse position into the position of the related character cell, return FALSE if the mouse pointer is out of the client rectangle's boundaries or if the console is not the foreground window
static BOOL CellCoordOfMousePos(const DWORD *const pAppBarMessage, const HWND hConWnd, const HANDLE hConOut, const PCOORD pCellCoord)
{
  POINT mousePos;
  RECT clientRect;
  CONSOLE_SCREEN_BUFFER_INFO bufferInfo;
  if (GetForegroundWindow() != hConWnd || // check if the console is the foreground window
      !GetCursorPos(&mousePos) || IsCursorOutOfWorkArea(pAppBarMessage, hConWnd, &mousePos) || // check if the mouse is outside of the work area of the monitor
      !ScreenToClient(hConWnd, &mousePos) || mousePos.x < 0 || mousePos.y < 0 || // convert screen coordinates of the mouse pointer into coordinates of the client rectangle
      !GetClientRect(hConWnd, &clientRect) || clientRect.right < mousePos.x || clientRect.bottom < mousePos.y || // check if the mouse is inside of the client rectangle
      !GetConsoleScreenBufferInfo(hConOut, &bufferInfo)) // get the boundaries of the client rectangle as indexes of character cells
    return FALSE;

  // best-fit calculation for the coordinates of the character cell beneath the mouse pointer
  pCellCoord->X = (SHORT)(mousePos.x / (LONG)(clientRect.right / (1.5 + bufferInfo.srWindow.Right - bufferInfo.srWindow.Left) + 0.5) + bufferInfo.srWindow.Left);
  pCellCoord->Y = (SHORT)(mousePos.y / (LONG)(clientRect.bottom / (1.5 + bufferInfo.srWindow.Bottom - bufferInfo.srWindow.Top) + 0.5) + bufferInfo.srWindow.Top);
  return TRUE;
}

// cache the color attribute of the current character cell, turn foreground into background color and background into foreground color
static void UpdateAttr(const HANDLE hConOut, const COORD *const pCellCoord, const PWORD pPrevAttr, const PCOORD pPrevCoord)
{
  WORD attr;
  if (!ReadConsoleOutputAttribute(hConOut, &attr, 1, *pCellCoord, (DWORD[]){ 0 }))
    return;

  *pPrevAttr = attr;
  attr = (attr & 0xFF00) | ((attr & 0x0F) << 4) | ((attr & 0xF0) >> 4); // swap background and foreground color
  WriteConsoleOutputAttribute(hConOut, &attr, 1, *pCellCoord, (DWORD[]){ 0 });
  *pPrevCoord = *pCellCoord;
}

// Defining this macro will only affect the GCC compiler and can reduce the size of the utility. As soon as you define it you have to add these linker options:
//   -Wl,-e,__start -nostartfiles
#define USE_START

#if defined(__GNUC__) && defined(USE_START)
# include <stdlib.h>
# define QUIT(_status) exit(_status)
# ifdef __i386
void _start(void);
void _start(void)
# else
void __start(void);
void __start(void)
# endif
#else
# define QUIT(_status) return (_status)
int main(void)
#endif
{
  // *** both launcher and worker process
  STARTUPINFOW startInfo;
  GetStartupInfoW(&startInfo);
  if (!(startInfo.dwFlags & STARTF_TITLEISAPPID)) // absence of STARTF_TITLEISAPPID identifies the launcher
  {
    // *** launcher process only
    startInfo.lpTitle = L"";
    startInfo.dwFlags = STARTF_TITLEISAPPID; // tag the worker process for identification
    PROCESS_INFORMATION procInfo;
    if (!CreateProcessW(NULL, GetCommandLineW(), NULL, NULL, FALSE, 0UL, NULL, NULL, &startInfo, &procInfo)) // start the worker process
      QUIT(1);

    CloseHandle(procInfo.hThread);
    CloseHandle(procInfo.hProcess);
    QUIT(0); // terminate the launcher process
  }

  // *** worker process only
  const HWND hConWnd = GetConsoleWindow(); // console window handle
  const HANDLE hConOut = CreateFileW(L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL); // handle of the output device of the console window
  const HCURSOR hollowCur = LoadCursorW(GetModuleHandleW(NULL), MAKEINTRESOURCEW(100)); // handle of the fully transparent mouse pointer which replaces the arrow pointer if the activated console is beneath the mouse
  DWORD shPid; // Process ID of the shell process
  HANDLE hShProc; // handle of the shell process which is running in the console window
  if (!hConWnd || hConOut == INVALID_HANDLE_VALUE || !hollowCur || !SendMessageW(hConWnd, WM_GETICON, 0, 0) || !(glob_defaultCur = CopyCursor(LoadCursorW(NULL, IDC_ARROW))) ||
      !SetConsoleCtrlHandler(RestoreDefaultCursor, TRUE) || !GetWindowThreadProcessId(hConWnd, &shPid) || !(hShProc = OpenProcess(SYNCHRONIZE, FALSE, shPid)))
    QUIT(1);

  const DWORD ver = GetVersion(); // this function is deprecated and will return the value for Windows 8 even on newer Windows versions, that's exactly good enough for our needs though
  const DWORD appBarMessage = (((ver & 0x00FF) << 8) | ((ver & 0xFF00) >> 8)) < _WIN32_WINNT_WIN8 ? ABM_GETAUTOHIDEBAR : 0x0000000B; // the latter is ABM_GETAUTOHIDEBAREX which is not defined before Win 8
  WORD prevAttr = 0; // cache for the original attributes of the updated cell
  COORD cellCoord = { 0 }, prevCoord = { 0 }; // caches for the current and previous cell position of the mouse pointer
  BOOL wasIn = FALSE; // marks whether the mouse pointer was previously inside the client rectangle of the console
  while (WaitForSingleObject(hShProc, 40) == WAIT_TIMEOUT) // polling; iterates as long as the shell process is alive; waits 40 ms to reduce CPU usage and to make the mouse pointer at least traceable when it is moved quickly
  {
    switch (CellCoordOfMousePos(&appBarMessage, hConWnd, hConOut, &cellCoord))
    {
    case FALSE: // the mouse pointer is outside of the client rectangle of the console
      if (wasIn && SetSystemCursor(CopyCursor(glob_defaultCur), OCR_NORMAL))
      {
        WriteConsoleOutputAttribute(hConOut, &prevAttr, 1, prevCoord, (DWORD[]){ 0 });
        wasIn = FALSE;
      }
      break;
    default: // the mouse pointer is inside of the client rectangle of the console
      switch (wasIn)
      {
      case FALSE: // the mouse pointer has previously been outside of the client rectangle of the console
        if (SetSystemCursor(CopyCursor(hollowCur), OCR_NORMAL))
        {
          UpdateAttr(hConOut, &cellCoord, &prevAttr, &prevCoord);
          wasIn = TRUE;
        }
        break;
      default: // the mouse pointer has previously been inside of the client rectangle of the console
        if (prevCoord.X != cellCoord.X || prevCoord.Y != cellCoord.Y)
        {
          WriteConsoleOutputAttribute(hConOut, &prevAttr, 1, prevCoord, (DWORD[]){ 0 });
          UpdateAttr(hConOut, &cellCoord, &prevAttr, &prevCoord);
        }
      }
    }
  }

  SetSystemCursor(glob_defaultCur, OCR_NORMAL);
  QUIT(0);
}

#ifdef _MSC_VER
# pragma warning(pop)
#endif
